#include "stdafx.h"
#include "AssetManager.h"
#include "Texture.h"
#include "Material.h"
#include "Mesh.h"
#include "MeshFlags.h"
#include "Slider.h"
#include <vector>

AssetManager* AssetManager::s_instance = 0;
const std::string s_texture_path = "data/Textures/";
const std::string s_mesh_path = "data/Meshes/";
const std::string s_material_path = "data/Materials/";

AssetManager::AssetManager()
{
	DBG_ASSERT(!s_instance);
	s_instance = this;
	m_offscrTexLoaded = false;
	m_currentProg = 0;
}

AssetManager::~AssetManager()
{
	DBG_ASSERT(s_instance);
	FreeAll();
	s_instance = 0;
}

void AssetManager::LoadAll()
{
	DBG_ASSERT(s_instance);
	LoadTexture("fire.png");
	LoadTexture("hud.png");

#if LOWPOLY
	LoadMesh("missile_lowpoly.obj");
	LoadMesh("spaceship_lowpoly.obj");
	LoadMesh("enemy2_lowpoly.obj");
#else
	LoadMesh("missile.obj", MeshFlags(true));
	LoadMesh("spaceship.obj", MeshFlags());
	LoadMesh("spaceship3.obj", MeshFlags());
	LoadMesh("enemy2.obj", MeshFlags(true));
#endif
	Mesh* splosion1 = LoadMesh("splosion1.obj", MeshFlags(true, true, true));
	Mesh* splosion2 = LoadMesh("splosion2.obj", MeshFlags(true, true, true));
	splosion1->AttachDeform(splosion2);

	LoadMesh("envCube.obj", MeshFlags());

	m_songPlayer.InitDevice();
	m_songPlayer.LoadFile((GetDataPath() + "ELSHOT.ogg").c_str());
	//m_songPlayer.LoadFile((GetDataPath() + "missile.obj").c_str());

	LoadMesh("fireball.obj", MeshFlags());
	LoadCubemapTexture("stormy", ".png");
	LoadCubemapTexture("hills", ".png");
	LoadCubemapTexture("mars", ".png");
	m_environmentMap.Load();

	// shaders
	m_instancedVS.Load("data/Shaders/Deform.glsl", GL_VERTEX_SHADER);
	m_defaultFS.Load("data/Shaders/FS.glsl", GL_FRAGMENT_SHADER);
	m_instanceShaderProg.Compile(m_instancedVS, m_defaultFS, true, false, false);

	m_simpleVS.Load("data/Shaders/SimpleVS.glsl", GL_VERTEX_SHADER);
	m_cubeMapFS.Load("data/Shaders/CubeMapFS.glsl", GL_FRAGMENT_SHADER);
	m_cubemapShaderProg.Compile(m_simpleVS, m_cubeMapFS, false, false, false);

	m_nonInstanceShaderProg.Compile(m_simpleVS, m_defaultFS, false, false, false);

	m_postProcessFS.Load("data/Shaders/PostProcessingFS.glsl", GL_FRAGMENT_SHADER);
	m_postProcessProg.Compile(m_simpleVS, m_postProcessFS, false, false, true);
	m_deformPostProcProg.Compile(m_instancedVS, m_postProcessFS, false, true, true);

	glGenFramebuffersEXT(1, &m_fbo);
	glGenRenderbuffersEXT(1, &m_depthBufferRB);
}

Texture* AssetManager::LoadTexture( std::string fileName )
{
	DBG_ASSERT(s_instance);

	
	// check if mesh already loaded
	DBG_ASSERT(!GetTexture(fileName, false));

	Texture* t = new Texture;
	bool ret;
	if(fileName == "OFFSCREEN") { //offscreen texture
		ret = t->Create(128, 128);
	}
	else { //load from file
		ret = t->Load(s_texture_path, fileName);
	}
	if(!ret) {
		DBG_ASSERT(0);
		return NULL;
	}

	m_textures.push_back(t);
	return t;
}

Mesh* AssetManager::LoadMesh( std::string fileName, MeshFlags flags)
{
	DBG_ASSERT(s_instance);
	
	// check if mesh already loaded
	DBG_ASSERT(!GetMesh(fileName, false));

	Mesh* m = new Mesh;

	// try to load from .bin
	std::ifstream fileStrm((GetDataPath() + fileName + ".bin").c_str(), std::ios_base::binary);
	if(fileStrm.good()) {
		m->LoadFromBin(fileStrm, flags);
		m->SetName(fileName);
	}
	else {
		bool ret = m->LoadObj(fileName, flags); // Load from obj instead
		DBG_ASSERT(ret);
		std::ofstream outFile((GetDataPath() + fileName + ".bin").c_str(), std::ios_base::binary);
		m->SaveToBin(outFile); // Save bin file
	}

	m_meshes.push_back(m);
	return m;
}

Texture* AssetManager::GetTexture( std::string name, bool assertOnFail)
{
	DBG_ASSERT(s_instance);
	for(int i=0; i<m_textures.size(); i++) {
		if(m_textures[i]->GetName() == name) return m_textures[i];
	}
	if(assertOnFail) DBG_ASSERT(false);
	return NULL;
}

Mesh* AssetManager::GetMesh( std::string name, bool assertOnFail)
{
	DBG_ASSERT(s_instance);
	for(int i=0; i<m_meshes.size(); i++) {
		if(m_meshes[i]->GetName() == name) 
			return m_meshes[i];
	}
	if(assertOnFail) DBG_ASSERT(false);
	return NULL;
}

Texture* AssetManager::GetOffscreenTexture(float width, float height)
{
	if(!m_offscrTexLoaded) {
		// offscreen texture
		m_offscreenTex.Create(width, height);
		m_offscrTexLoaded = true;
	}
	if(m_offscreenTex.GetWidth() != width || m_offscreenTex.GetHeight() != height)
	{
		m_offscreenTex.Free();
		m_offscreenTex.Create(width, height);
		m_offscrTexLoaded = true;
	}
	return &m_offscreenTex;
}

Material AssetManager::GetMaterial( std::string name, bool assertOnFail)
{
	DBG_ASSERT(s_instance);
	for(int i=0; i<m_materials.size(); i++) {
		if(m_materials[i].GetName() == name) {
			return m_materials[i];
		}
	}
	if(assertOnFail) DBG_ASSERT(false);
	return Material();
}

void AssetManager::FreeAll()
{
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	glDeleteFramebuffersEXT(1, &m_fbo);
	//glDeleteRenderbuffersEXT(1, &m_depthbuffer);

	DBG_ASSERT(s_instance);
	while(!m_textures.empty()) {
		Texture* t = *(m_textures.end()-1);
		t->Free();
		m_textures.pop_back();
		delete t;
	}
	while(!m_meshes.empty()) {
		Mesh* m = *(m_meshes.end()-1);
		m_meshes.pop_back();
		delete m;
	}
}

Texture* AssetManager::LoadCubemapTexture( std::string prefix, std::string extension)
{
	DBG_ASSERT(s_instance);

	// check if mesh already loaded
	DBG_ASSERT(!GetTexture(prefix, false));

	Texture* t = new Texture;
	bool ret = t->LoadCubemap(prefix, extension);
	if(!ret) {
		DBG_ASSERT(0);
		return NULL;
	}

	m_textures.push_back(t);
	return t;
	
}

void AssetManager::LoadMtllib( std::string fileName )
{
	std::ifstream fileStream((GetDataPath()+fileName).c_str());
	if(!fileStream.good()) DBG_ASSERT(0);

	Material m;
	bool firstMaterial = true;

	while (fileStream.good())
	{
		std::string tag;
		fileStream >> tag;

		if(!tag.compare("newmtl")) {
			if(!firstMaterial) {
				m_materials.push_back(m);
				m = Material();
			}
			fileStream >> m.m_name;
			if(m.m_name == "Fire") m.m_type = FIRE;
			else if(m.m_name == "Metal") m.m_type = METAL;
			else if(m.m_name == "Glass") m.m_type = GLASS;
			firstMaterial = false;
		}
		else if(!tag.compare("Kd")) {
			fileStream >> m.m_col[0] >> m.m_col[1] >> m.m_col[2];			
		}
		else if(!tag.compare("#")) {
			// comment, ignore line
			std::string comment;
			getline(fileStream, comment);
		}
	}
	m_materials.push_back(m);
}

void AssetManager::UseProgram( ShaderProgram* p )
{
	if(!p) {
		glUseProgram(0);
		m_currentProg = 0;
	}
	else {
		glUseProgram(p->GetId()); 
		m_currentProg=p;
	}
}

void AssetManager::AddParameter(std::string name, unsigned int* i, unsigned int lo, unsigned int hi)
{
	Slider s(name, i, lo, hi);
	m_params.push_back(s);
}
void AssetManager::AddParameter(std::string name, float* f)
{
	Slider s(name, f);
	m_params.push_back(s);
}