#include "stdafx.h"
#include "AssetManager.h"
#include "Glyphs.h"
#include "cairo/cairo.h"
#include <map>

bool s_loaded = false;
unsigned int s_glTexture = 0;
float s_glyphImgWidth = 0;
float s_glyphImgHeight = 0;
float s_textHeight = 0;
std::map<char, GlyphBounds> s_glyphBounds;

unsigned int Glyphs::GetTexture()
{
	return s_glTexture;
}

void Glyphs::CreateTexture()
{
	glEnable(GL_TEXTURE_2D);
	glEnable (GL_BLEND);
	glGenTextures(1, &s_glTexture);
	glBindTexture(GL_TEXTURE_2D, s_glTexture);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND); //allow transparency

	const int glyphImgSize = 300;
	s_glyphImgWidth = glyphImgSize;
	s_glyphImgHeight = glyphImgSize;

	// create a cairo surface
	cairo_surface_t *surface;
	cairo_t *cr;
	surface = cairo_image_surface_create (CAIRO_FORMAT_RGB24, glyphImgSize, glyphImgSize);
	cr = cairo_create (surface);

	// set font properties
	const float FONTSIZE = 16;
	cairo_select_font_face (cr, "serif", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_BOLD);
	cairo_set_font_size (cr, FONTSIZE);
	cairo_set_source_rgb (cr, 1.0, 1.0, 1.0);

	cairo_font_options_t* fontOptions = cairo_font_options_create();
	cairo_get_font_options(cr, fontOptions);
	cairo_font_options_set_antialias(fontOptions, CAIRO_ANTIALIAS_NONE);
	cairo_set_font_options(cr, fontOptions);
	cairo_font_options_destroy(fontOptions);

	double x = 0;
	double y = glyphImgSize;
	double line_height = 0;
	cairo_font_extents_t font_extents;
	cairo_font_extents(cr, &font_extents);
	cairo_move_to (cr, x, y);
	for(int i = 32; i < 126; i++) {
		//draw to cairo surf
		char str[2];
		str[0] = (char)i;
		str[1] = '\0';

		// Blit character
		const float SCALE = 0.5;
		cairo_text_path (cr, str);
		cairo_set_source_rgb (cr, 1, 1, 1); // the glyph itself
		cairo_fill_preserve (cr);
		//cairo_set_source_rgb (cr, 0.0, 0.0, 1.0); // outline
		//cairo_set_line_width (cr, FONTSIZE*0.04);
		//cairo_stroke (cr);
		cairo_close_path (cr);
		cairo_fill (cr);

		// Add to list
		cairo_text_extents_t extents;
		cairo_text_extents(cr, str, &extents);
		double descent = extents.y_bearing;
		
		s_glyphBounds[str[0]] = GlyphBounds(x, y+font_extents.descent-3, extents.x_advance, font_extents.height);
		
		x += extents.x_advance;
		if(x + extents.width > glyphImgSize) { // wrap characters
			x = 0;
			y -= font_extents.height;
		}
		cairo_move_to (cr, x, y);
	}
	DBG_ASSERT(x < glyphImgSize && y < glyphImgSize);
	//cairo_surface_write_to_png (surface, "glyphs.png");
	s_textHeight = font_extents.height;

	// allocate gl texture
	glTexImage2D(GL_TEXTURE_2D, 0, 4, glyphImgSize, glyphImgSize, 0, GL_RGBA, GL_UNSIGNED_BYTE, cairo_image_surface_get_data(surface));

	// free cairo surface
	cairo_destroy (cr);
	cairo_surface_destroy (surface);
	s_loaded = true;
}
void Glyphs::ShowText( std::string label, double x, double y, double maxx, double maxy, bool ralign )
{
	if(!s_loaded) return;

	glActiveTexture(GL_TEXTURE0 + 0);
	glBindTexture(GL_TEXTURE_2D, s_glTexture);
	glEnable(GL_TEXTURE_2D);

	float xDisp = 0;
	if(ralign) {
		// measure text
		for(int i=0; i<label.size(); i++) {
			xDisp += s_glyphBounds[label[i]].width;
		}
	}

	glPushMatrix();
	glTranslated(x-xDisp, y-(s_textHeight/2), 0);
	for(int i=0; i<label.size(); i++) {
		// for each glyphbound, (x, y) is the lower left corner of the glyph on the bitmap.
		// the upper right corner is (x+width, y-height).
		char ch = label[i]; // just draw 1st character for now
		DBG_ASSERT(s_glyphBounds.find(ch) != s_glyphBounds.end());
		GlyphBounds bounds = s_glyphBounds[ch];

		// texcoords are scaled to [0, 1]
		double tx1 = bounds.x/s_glyphImgWidth;
		double ty1 = bounds.y/s_glyphImgHeight;
		double tx2 = (bounds.x+bounds.width)/s_glyphImgWidth;
		double ty2 = (bounds.y-bounds.height)/s_glyphImgHeight;

		glBegin(GL_QUADS);
		glTexCoord2d(tx1, ty1);
		glVertex2d(0, 0);
		glTexCoord2d(tx1, ty2);
		glVertex2d(0, bounds.height);
		glTexCoord2d(tx2, ty2);
		glVertex2d(bounds.width, bounds.height);
		glTexCoord2d(tx2, ty1);
		glVertex2d(bounds.width, 0);
		glEnd();

		glTranslated(bounds.width, 0, 0);
	}
	glPopMatrix();
	
	// unbind texture
	glActiveTexture(GL_TEXTURE0 + 0);
	glBindTexture(GL_TEXTURE_2D, 0);
	glDisable(GL_TEXTURE_2D);
}

void Glyphs::Free()
{
	glDeleteTextures(1, &s_glTexture);
	s_loaded = false;
}