#include "stdafx.h"
#include "Mesh.h"
#include "AssetManager.h"
#include "MathUtils.h"

struct OBJFile
{
	std::vector<Point3D> positions;
	std::vector<Vector3D> normals;
	std::vector<Point2D> texCoords;
	Vertex LookupVertex(VertexIndices inds)
	{
		Vertex v;
		v.position = positions[inds.v];
		if(inds.useVt) v.texCoord = texCoords[inds.vt];
		if(inds.useVn) v.normal = normals[inds.vn];
		return v;
	}
};

const int Mesh::END_OF_MESH = 28;

void Mesh::LoadFromBin(std::ifstream& fileStrm, MeshFlags flags)
{
	DBG_ASSERT(m_subMeshes.size()==0);
	unsigned int numSubmeshes;
	fileStrm.read((char*)&numSubmeshes, sizeof(unsigned int));
	for(int i=0; i<numSubmeshes; i++) {
		SubMesh s;
		s.LoadFromBin(fileStrm, flags);
		m_subMeshes.push_back(s);
	}
	int flag;
	fileStrm.read((char*)&flag, sizeof(int));
	DBG_ASSERT(flag == END_OF_MESH);

	// calc bounds
	CalcBoundingBox();

	m_loaded = true;
};

void Mesh::SaveToBin(std::ofstream& fileStrm)
{
	unsigned int numSubMeshes = m_subMeshes.size();
	fileStrm.write((char*)&numSubMeshes, sizeof(unsigned int));
	for(int i=0; i<m_subMeshes.size(); i++) {
		m_subMeshes[i].SaveToBin(fileStrm);
	}
	fileStrm.write((char*)&END_OF_MESH, sizeof(int));
	fileStrm.close();
};

Mesh::Mesh()
{
	m_loaded = false;
}

bool Mesh::LoadObj(std::string fileName, MeshFlags flags)
{
	std::ifstream fileStream((AssetManager::GetInstance()->GetDataPath() + fileName).c_str());
	if(!fileStream.good()) return false;

	OBJFile objFile;

	std::vector<std::vector<VertexIndices> > faces;
	bool firstSubmesh = true;
	SubMesh subMesh;

	while (fileStream.good())
	{
		std::string tag;
		fileStream >> tag;

		if(!tag.compare("g")) {
			// make everything apply to a new submesh
			if(!firstSubmesh) {
				subMesh.SetName(fileName);
				subMesh.BuildVBOs();
				subMesh.CalcBoundingBox();
				m_subMeshes.push_back(subMesh);
			}
			firstSubmesh = false;
			subMesh = SubMesh();
			std::string name;
			fileStream >> name;
			subMesh.SetTitle(name);
		}
		else if(!tag.compare("usemtl")) {
			std::string matName;
			fileStream >> matName;
			//subMesh.SetMaterial(AssetManager::GetInstance()->GetMaterial(matName));
		}
		else if(!tag.compare("f")) {
			std::string indexList;
			getline(fileStream, indexList);
			std::istringstream indexListStrm(indexList);
			std::vector<VertexIndices> face;
			while(1) {
				VertexIndices indices;
				indexListStrm >> indices;
				if(!indexListStrm.good()) break;
				face.push_back(indices);
			}
			// convert quad to triangles
			if(face.size()==4) {
				std::vector<VertexIndices> newface1, newface2;
				Vertex t1[3] = {objFile.LookupVertex(face[0]), objFile.LookupVertex(face[2]), objFile.LookupVertex(face[1])};
				Vertex t2[3] = {objFile.LookupVertex(face[0]), objFile.LookupVertex(face[3]), objFile.LookupVertex(face[2])};
				subMesh.AddTriangle(t1[0], t1[1], t1[2]);
				subMesh.AddTriangle(t2[0], t2[1], t2[2]);
			}
			else if(face.size()==3) {				
				subMesh.AddTriangle(objFile.LookupVertex(face[0]), objFile.LookupVertex(face[1]), objFile.LookupVertex(face[2]));
			}
			else {
				DBG_ASSERT(0);
			}
			// add face to current submesh
		}
		else if(!tag.compare("mtllib")) {
			std::string mtllibList;
			getline(fileStream, mtllibList);
			std::istringstream mtllibListStrm(mtllibList);
			while(1) {
				std::string libName;
				mtllibListStrm >> libName;
				//AssetManager::GetInstance()->LoadMtllib(libName);
				if(!mtllibListStrm.good()) break;			
			}
		}
		else if(!tag.compare("v")) {
			double v[3];
			fileStream >> v[0] >> v[1] >> v[2];
			objFile.positions.push_back(Point3D(v[0], v[1], v[2]));
		}
		else if(!tag.compare("vn")) {
			double vn[3];
			fileStream >> vn[0] >> vn[1] >> vn[2];
			objFile.normals.push_back(Vector3D(vn[0], vn[1], vn[2]));
		}
		else if(!tag.compare("vt")) {
			double vt[2];
			fileStream >> vt[0] >> vt[1];
			objFile.texCoords.push_back(Point2D(vt[0], vt[1]));
		}
	}
	subMesh.SetName(fileName);
	subMesh.BuildVBOs();
	subMesh.CalcBoundingBox();
	m_subMeshes.push_back(subMesh);

	CalcBoundingBox();

	m_name = fileName;
	m_loaded = true;
	return true;
}

float Mesh::GetExtent( int dim )
{
	DBG_ASSERT(dim >=0 && dim <=3 && m_upperBound[dim] > m_lowerBound[dim]);
	return m_upperBound[dim] - m_lowerBound[dim];
}

void Mesh::CalcBoundingBox()
{
	if(m_subMeshes.size()==0) return;
	m_lowerBound = m_subMeshes[0].GetLowerBound();
	m_upperBound = m_subMeshes[0].GetUpperBound();
	m_maxExtent = 0;
	for(int i=1; i<m_subMeshes.size(); i++) {
		for(int dim=0; dim<3; dim++) {
			m_lowerBound[dim] = MATH_MIN(m_lowerBound[dim], m_subMeshes[i].GetLowerBound()[dim]);
			m_upperBound[dim] = MATH_MAX(m_upperBound[dim], m_subMeshes[i].GetUpperBound()[dim]);
		}
	}
	for(int dim=0; dim<3; dim++) {
		m_maxExtent = MATH_MAX(m_maxExtent, m_upperBound[dim] - m_lowerBound[dim]);
	}
}

void Mesh::UpdateDeform(float t)
{
	for(int i=0; i<m_subMeshes.size(); i++) {
		m_subMeshes[i].UpdateDeform(t);
	}
}

void Mesh::DrawInstanced(std::vector<Matrix4x4> transforms)
{
	for(int i=0; i<m_subMeshes.size(); i++) {
		m_subMeshes[i].DrawInstanced(transforms);
	}
}

void Mesh::SetInstancing(Matrix4x4 m, int index)
{
	if(!m_loaded) return;
	for(int i=0; i<m_subMeshes.size(); i++) {
		m_subMeshes[i].SetInstancing(m, index);
	}
}

void Mesh::SetInstancing( std::vector<Matrix4x4> transforms )
{
	for(int i=0; i<m_subMeshes.size(); i++) {
		m_subMeshes[i].SetInstancing(transforms);
	}
}

void Mesh::SetInstancing( std::vector<Matrix4x4> transforms, unsigned int startInd, unsigned int count )
{
	for(int i=0; i<m_subMeshes.size(); i++) {
		m_subMeshes[i].SetInstancing(transforms, startInd, count);
	}
}

void Mesh::Draw(unsigned int numInstances)
{
	if(numInstances==0) return;
	for(int i=0; i<m_subMeshes.size(); i++) {
		if(m_subMeshes[i].GetMaterial().IsFire()) continue;
		m_subMeshes[i].ApplyMaterial();
		m_subMeshes[i].Draw(numInstances);
	}
}

void Mesh::DrawFires(bool forceMat)
{
	for(int i=0; i<m_subMeshes.size(); i++) {
		if(m_subMeshes[i].GetMaterial().IsFire() || forceMat) {
			m_subMeshes[i].DrawFire(forceMat);
		}
	}
}

void Mesh::DrawCubemap(Texture* t1, Texture* t2, float transition)
{
	DBG_ASSERT(m_subMeshes.size()==1);
	for(int i=0; i<m_subMeshes.size(); i++) {
		m_subMeshes[i].DrawCubemap(t1, t2, transition);
	}
}

void Mesh::AttachDeform( Mesh* m )
{
	DBG_ASSERT(m_subMeshes.size() == m->GetNumSubMeshes());
	for(int i=0; i<m_subMeshes.size(); i++) {
		m_subMeshes[i].AttachDeform(m->GetSubMesh(i));
	}	
}