#include "stdafx.h"
#include "SubMesh.h"
#include "MeshFlags.h"
#include "MathUtils.h"
#include "AssetManager.h"
#include "Shader.h"
#include "ShaderProgram.h"
#include "Texture.h"
#include "TripletMap.h"
#include "algebra.h"
#include <map>
#include <iostream>

#define MAX_UINT 0xFFFFFFFF

Point3D SubMesh::m_ambient(0.8, 0.8, 0.8);
const int SubMesh::END_OF_SUBMESH = 27;

void SubMesh::SwapYZ()
{
	for(int i=0; i<m_vertPosList.size(); i++) {
		double tmp = m_vertPosList[i].v_[1];
		m_vertPosList[i].v_[1] = m_vertPosList[i].v_[2];
		m_vertPosList[i].v_[2] = tmp;
	}
}

void SubMesh::BuildVBOs()
{
	glGenBuffersARB(1, &m_vertexVbo);
	glGenBuffersARB(1, &m_vertexIndexBuf);
	glGenBuffersARB(1, &m_normalVbo);
	glGenBuffers(1, &m_instanceIdBuf);
	glGenBuffersARB(1, &m_texCoordVbo);	

	glBindBufferARB(GL_ARRAY_BUFFER_ARB, m_vertexVbo);
	//load vertex data
	m_positionBuf = new float[m_vertPosList.size()*3*MAX_INSTANCES];
	int arrInd = 0;
	for (int inst=0; inst<MAX_INSTANCES; inst++)
	{
		for(int i=0; i<m_vertPosList.size(); i++) {
			m_positionBuf[arrInd] = m_vertPosList[i].v_[0];
			m_positionBuf[arrInd+1] = m_vertPosList[i].v_[1];
			m_positionBuf[arrInd+2] = m_vertPosList[i].v_[2];
			arrInd+=3;
		}
	}
	glBufferDataARB( GL_ARRAY_BUFFER_ARB, m_vertPosList.size()*3*MAX_INSTANCES*sizeof(float), m_positionBuf, GL_STATIC_DRAW_ARB);

	glBindBufferARB(GL_ARRAY_BUFFER_ARB, m_normalVbo);
	//load normal data
	DBG_ASSERT(m_vertPosList.size() == m_normalList.size());
	m_normalBuf = new float[m_normalList.size()*3*MAX_INSTANCES];
	arrInd = 0;
	for (int inst=0; inst<MAX_INSTANCES; inst++)
	{
		for(int i=0; i<m_normalList.size(); i++) {
			DBG_ASSERT(arrInd < m_normalList.size()*3*MAX_INSTANCES);
			m_normalBuf[arrInd] = m_normalList[i][0];	
			m_normalBuf[arrInd+1] = m_normalList[i][1];
			m_normalBuf[arrInd+2] = m_normalList[i][2];
			arrInd+=3;
		}
	}
	glBufferDataARB( GL_ARRAY_BUFFER_ARB, m_normalList.size()*3*MAX_INSTANCES*sizeof(float), m_normalBuf, GL_STATIC_DRAW_ARB);

	glBindBufferARB(GL_ARRAY_BUFFER_ARB, m_texCoordVbo);
	//load tex coord data
	DBG_ASSERT(m_vertPosList.size() == m_texCoordList.size());
	float* texCoords = new float[m_texCoordList.size()*2*MAX_INSTANCES];
	arrInd = 0;
	for (int inst=0; inst<MAX_INSTANCES; inst++)
	{
		for(int i=0; i<m_texCoordList.size(); i++) {
			texCoords[arrInd] = m_texCoordList[i][0];
			texCoords[arrInd+1] = m_texCoordList[i][1];
			arrInd+=2;
		}
	}
	glBufferDataARB( GL_ARRAY_BUFFER_ARB, m_texCoordList.size()*2*MAX_INSTANCES*sizeof(float), texCoords, GL_STATIC_DRAW_ARB);
	delete [] texCoords;

	//create instanceId buffer.
	glBindBuffer(GL_ARRAY_BUFFER, m_instanceIdBuf);
	float* instanceIds = new float[m_vertPosList.size()*MAX_INSTANCES];
	for(int inst=0; inst<MAX_INSTANCES; inst++) {
		for(int i=0; i<m_vertPosList.size(); i++) {
			DBG_ASSERT(m_vertPosList.size()*inst+i < m_vertPosList.size()*MAX_INSTANCES);
			instanceIds[m_vertPosList.size()*inst+i] = inst;
		}
	}
	glBufferDataARB( GL_ARRAY_BUFFER_ARB, m_vertPosList.size()*MAX_INSTANCES*sizeof(float), instanceIds, GL_STATIC_DRAW_ARB);
	delete [] instanceIds;

	m_loaded = true;
}

std::istream &operator>>(std::istream &stream, VertexIndices &ob)
{ 
	//read each token in a stream, i.e. a token is v, v//vn, v/vt, or v/vt/vn.
	ob.v = -1;
	ob.vn = -1;
	ob.vt = -1;

	std::string token;
	stream >> token;
	size_t lhslash = token.find('/');
	size_t rhslash = token.rfind('/');
	std::string vStr = token.substr(0, lhslash);
	std::istringstream vStrm(vStr);
	vStrm >> ob.v;
	ob.v--;
	if(lhslash == std::string::npos) {ob.useVn = false; ob.useVt=false;} // format is v
	else if(lhslash + 1 == rhslash) { // format is v//vn
		std::string vnStr = token.substr(rhslash+1, token.size());
		std::istringstream vnStrm(vnStr);
		vnStrm >> ob.vn;
		ob.vn--; // use 0-indexing
		ob.useVt = false;
	}
	else if(lhslash == rhslash) { // v/vt
		std::string vtStr = token.substr(rhslash+1, token.size());
		std::istringstream vtStrm(vtStr);
		vtStrm >> ob.vt;
		ob.vt--;
		ob.useVn = false;
	}
	else { // v/vt/vn.
		std::string vtStr = token.substr(lhslash+1, rhslash);
		std::istringstream vtStrm(vtStr);
		vtStrm >> ob.vt;
		ob.vt--;
		std::string vnStr = token.substr(rhslash+1, token.size());
		std::istringstream vnStrm(vnStr);
		vnStrm >> ob.vn;
		ob.vn--;
	}
	return stream;
}

bool SubMesh::LoadObj(std::string fileName, MeshFlags flags)
{
	m_flags = flags;
	std::ifstream fileStream((AssetManager::GetInstance()->GetDataPath()+fileName).c_str());
	if(!fileStream.good()) return false;

	while (fileStream.good())
	{
		std::string tag;
		fileStream >> tag;

		if(!tag.compare("f")) {
			std::string indexList;
			getline(fileStream, indexList);
			std::istringstream indexListStrm(indexList);
			std::vector<VertexIndices> face;
			while(1) {
				VertexIndices indices;
				indexListStrm >> indices;
				if(!indexListStrm.good()) break;
				face.push_back(indices);
			}
			// convert quad to triangles
			if(face.size()==4) {
				std::vector<VertexIndices> newface1, newface2;
				newface1.push_back(face[0]);
				newface1.push_back(face[2]);
				newface1.push_back(face[1]);
				newface2.push_back(face[0]);
				newface2.push_back(face[3]);
				newface2.push_back(face[2]);
				m_faces.push_back(newface1);
				m_faces.push_back(newface2);
			}
			else if(face.size()==3) {
				m_faces.push_back(face);
			}
			else {
				DBG_ASSERT(0);
			}
		}
		else if(!tag.compare("mtllib")) {
			std::string workingDir;
			size_t ind1 = fileName.rfind('/');
			size_t ind2 = fileName.rfind('\\');
			if(ind1 < fileName.size()-1) workingDir += fileName.substr(0, ind1+1);
			else if(ind2 < fileName.size()-1) workingDir += fileName.substr(0, ind2+1);
			std::string mtllibList;
			getline(fileStream, mtllibList);
			std::istringstream mtllibListStrm(mtllibList);
			while(1) {
				std::string libName;
				mtllibListStrm >> libName;
				m_material = AssetManager::GetInstance()->GetMaterial(libName, false);	
				if(!mtllibListStrm.good()) break;			
			}
		}
		else if(!tag.compare("v")) {
			double v[3];
			fileStream >> v[0] >> v[1] >> v[2];
			m_vertPosList.push_back(Point3D(v[0], v[1], v[2]));
		}
		else if(!tag.compare("vn")) {
			double vn[3];
			fileStream >> vn[0] >> vn[1] >> vn[2];
			m_normalList.push_back(Vector3D(vn[0], vn[1], vn[2]));
		}
		else if(!tag.compare("vt")) {
			double vt[2];
			fileStream >> vt[0] >> vt[1];
			m_texCoordList.push_back(Point2D(vt[0], vt[1]));
		}
	}

	BuildVerticesFromRaw();
	BuildVBOs();
	CalcBoundingBox();
	
	m_name = fileName;

	m_loaded = true;
	return true;
}

void SubMesh::SetMaterial(Material m)
{
	m_material = m;
}

void SubMesh::DrawInstanced( std::vector<Matrix4x4> transforms )
{
	// Make draw calls in batches of MAX_INSTANCES
	int instLeft = transforms.size();
	int batchNum = 0;
	while(instLeft>0) {
		int numToDraw = MATH_MIN(MAX_INSTANCES, instLeft);
		SetInstancing(transforms, batchNum*MAX_INSTANCES, numToDraw);
		Draw(numToDraw);
		instLeft -= numToDraw;
		batchNum++;
	}
}

void SubMesh::ApplyMaterial()
{
	if(m_material.m_type == METAL) { // leave defaults
	}
}

void SubMesh::DrawFire(bool forceMat)
{
	if(!m_loaded) return;

	SetDeformDrawParams();
	SetInstanceDrawParams();

	glBindBuffer(GL_ARRAY_BUFFER, m_vertexVbo);         // for vertex coordinates
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glVertexPointer(3, GL_FLOAT, 0, 0);
	glBindBuffer(GL_ARRAY_BUFFER, m_normalVbo);
	glNormalPointer(GL_FLOAT, 0, 0);
	glBindBuffer(GL_ARRAY_BUFFER, m_texCoordVbo);
	glTexCoordPointer(2, GL_FLOAT, 0, 0);

	// draw
	glDrawArrays(GL_TRIANGLES, 0, m_numFaces*3);

	glDisableClientState(GL_VERTEX_ARRAY); // deactivate vertex array

	// bind with 0, so, switch back to normal pointer operation
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
}

void SubMesh::Draw(int numInstances)
{
	if(!m_loaded) return;
	if(numInstances>1 && !m_flags.useInstancing) DBG_ASSERT(0);

	SetDeformDrawParams();
	SetInstanceDrawParams();
	
	
	// bind fire texture
	//glUniform1i(AssetManager::GetInstance()->GetCurrentProgram()->GetVarId("fireTex"), 1);
	
	// cubemap coords, if reflective
	glUniform1i(AssetManager::GetInstance()->GetCurrentProgram()->GetVarId("cubeTex"), 0);
	glActiveTexture(GL_TEXTURE0);
	glEnable(GL_TEXTURE_CUBE_MAP);
	glBindTexture(GL_TEXTURE_CUBE_MAP, AssetManager::GetInstance()->GetEnvironmentMap()->GetTexture()->GetId());
	glEnable(GL_TEXTURE_GEN_S); 
	glEnable(GL_TEXTURE_GEN_T); 
	glEnable(GL_TEXTURE_GEN_R);
	glEnable(GL_NORMALIZE);
	
	glBindBuffer(GL_ARRAY_BUFFER, m_vertexVbo);         // for vertex coordinates
	//glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_vertexIndexBuf);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glVertexPointer(3, GL_FLOAT, 0, 0);
	glBindBuffer(GL_ARRAY_BUFFER, m_normalVbo);
	glNormalPointer(GL_FLOAT, 0, 0);
	glBindBuffer(GL_ARRAY_BUFFER, m_texCoordVbo);
	glTexCoordPointer(2, GL_FLOAT, 0, 0);
	
	// draw
	//glDrawElements(GL_TRIANGLES, m_numFaces*3*numInstances, GL_UNSIGNED_INT, 0);
	glDrawArrays(GL_TRIANGLES, 0, m_numFaces*3*numInstances);

	glDisableClientState(GL_VERTEX_ARRAY); // deactivate vertex array

	// bind with 0, so, switch back to normal pointer operation
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
		
}
bool SubMesh::AttachDeform(SubMesh* m)
{
	//generate buffer for deform info
	glGenBuffers(1, &m_deformInfo.m_deformPosBuf);
	glGenBuffers(1, &m_deformInfo.m_deformNormBuf);
	glGenBuffers(1, &m_deformInfo.m_deformSpeedBuf);

	SubMesh* deformMesh = m;
	//make sure other mesh has same num of vertices/indices. should have same indices also
	DBG_ASSERT(GetNumVertices() == deformMesh->GetNumVertices());
	DBG_ASSERT(GetNumIndices() == deformMesh->GetNumIndices());
	DBG_ASSERT(GetNumNormals() == deformMesh->GetNumNormals());
	
	glBindBuffer(GL_ARRAY_BUFFER, m_deformInfo.m_deformPosBuf);
	glBufferData(GL_ARRAY_BUFFER, deformMesh->GetNumVertices()*3*MAX_INSTANCES*sizeof(float), deformMesh->GetPositionBuffer(), GL_STATIC_DRAW);

	glBindBuffer(GL_ARRAY_BUFFER, m_deformInfo.m_deformNormBuf);
	glBufferData(GL_ARRAY_BUFFER, deformMesh->GetNumNormals()*3*MAX_INSTANCES*sizeof(float), deformMesh->GetNormalBuffer(), GL_STATIC_DRAW);

	//glBindBuffer(GL_ARRAY_BUFFER, m_deformInfo.m_deformSpeedBuf);
	//glBufferData(GL_ARRAY_BUFFER, deformMesh->GetNumVertices()*MAX_INSTANCES*sizeof(float), speeds, GL_STATIC_DRAW);
	//delete [] speeds;
	m_flags.useDeform = true;

	m_deformInfo.m_percent = 0;
	return true;
}

void SubMesh::UpdateDeform(double percent)
{
	DBG_ASSERT(m_flags.useDeform);
	if(percent < 0) percent = 0;
	if(percent > 1) percent = 1;
	m_deformInfo.m_percent = percent;
	//m_deformInfo.m_percent = 0;
}

Vector3D CalcNormal(Point3D p1, Point3D p2, Point3D p3)
{
	Vector3D a = p2 - p1;
	Vector3D b = p3 - p1;
	return a.cross(b);
}

void SubMesh::BuildVerticesFromRaw()
{
	std::vector<Point3D> newVertPosList;
	std::vector<Point2D> newTexCoordList;
	std::vector<Vector3D> newNormalList;
	TripletMap coordMap;

	for(int i=0; i<m_faces.size(); i++) {
		for(int j=0; j<3; j++) {
			if(false && coordMap.Contains(m_faces[i][j].v, m_faces[i][j].vt, m_faces[i][j].vn)) {
				unsigned int index = coordMap.Lookup(m_faces[i][j].v, m_faces[i][j].vt, m_faces[i][j].vn);
				m_indexList.push_back(index);

			}
			else {
				coordMap.Insert(m_faces[i][j].v, m_faces[i][j].vt, m_faces[i][j].vn, m_indexList.size());
				Point3D pos = m_vertPosList[m_faces[i][j].v];
				Point2D texCoord = m_texCoordList[m_faces[i][j].vt];
				Vector3D normal = m_normalList[m_faces[i][j].vn];
				newVertPosList.push_back(pos);
				newTexCoordList.push_back(texCoord);
				newNormalList.push_back(normal);
				m_indexList.push_back(m_indexList.size());
			}
		}
		
	}
	//m_vertPosList.clear();
	//m_texCoordList.clear();
	//m_normalList.clear();
	m_vertPosList = std::vector<Point3D>(newVertPosList);
	m_texCoordList = std::vector<Point2D>(newTexCoordList);
	m_normalList = std::vector<Vector3D>(newNormalList);
}

void SubMesh::CalcBoundingBox()
{
	DBG_ASSERT(m_vertPosList.size()>0);
	m_lowerBound = m_vertPosList[0];
	m_upperBound = m_vertPosList[0];
	for(int i=1; i<m_vertPosList.size(); i++) {
		for(int j=0; j<3; j++) {
			m_lowerBound[j] = MATH_MIN(m_lowerBound[j], m_vertPosList[i][j]);
			m_upperBound[j] = MATH_MAX(m_upperBound[j], m_vertPosList[i][j]);
		}
	}
	m_maxExtent = 0;
	for(int i=0; i<3; i++) {
		m_maxExtent = MATH_MAX(m_maxExtent, m_upperBound[i] - m_lowerBound[i]);
	}
}

void SubMesh::SetDeformDrawParams()
{
	if(!m_flags.useInstancing) return;
	// set dest positions
	if(m_flags.useDeform) glBindBuffer(GL_ARRAY_BUFFER, m_deformInfo.m_deformPosBuf);
	else glBindBuffer(GL_ARRAY_BUFFER, m_vertexVbo);
	// setup position2 buffer. only used by deform shader..
	glEnableVertexAttribArray(AssetManager::GetInstance()->GetCurrentProgram()->GetVarId("position2"));
	glVertexAttribPointer(AssetManager::GetInstance()->GetCurrentProgram()->GetVarId("position2"), 3, GL_FLOAT, GL_FALSE, 0, 0);

	// dest normals
	if(m_flags.useDeform) glBindBuffer(GL_ARRAY_BUFFER, m_deformInfo.m_deformNormBuf);
	else glBindBuffer(GL_ARRAY_BUFFER, m_normalVbo);
	glEnableVertexAttribArray(AssetManager::GetInstance()->GetCurrentProgram()->GetVarId("normal2"));
	glVertexAttribPointer(AssetManager::GetInstance()->GetCurrentProgram()->GetVarId("normal2"), 3, GL_FLOAT, GL_FALSE, 0, 0);

	/*
	// speeds
	glBindBuffer(GL_ARRAY_BUFFER, m_deformInfo.m_deformSpeedBuf);
	glEnableVertexAttribArray(m_deformSpeedVar);
	glVertexAttribPointer(m_deformSpeedVar, 1, GL_FLOAT, GL_FALSE, 0, 0);*/

	// set deform parameter
	if(m_flags.useDeform) {
		glUniform1f(AssetManager::GetInstance()->GetCurrentProgram()->GetVarId("percent"), m_deformInfo.m_percent);
	}
	else {
		glUniform1f(AssetManager::GetInstance()->GetCurrentProgram()->GetVarId("percent"), 0);
	}
}

void SubMesh::SetInstanceDrawParams()
{
	if(m_flags.useInstancing) {
		glBindBuffer(GL_ARRAY_BUFFER, m_instanceIdBuf);
		glEnableVertexAttribArray(AssetManager::GetInstance()->GetCurrentProgram()->GetVarId("instanceId"));
		glVertexAttribPointer(AssetManager::GetInstance()->GetCurrentProgram()->GetVarId("instanceId"), 1, GL_FLOAT, GL_FALSE, 0, 0); //1-component int
		glUniformMatrix4fv(AssetManager::GetInstance()->GetCurrentProgram()->GetVarId("instanceTransMats"), MAX_INSTANCES, false, m_instanceData.GetBuffer());
	}
}

void SubMesh::SetInstancing(Matrix4x4 m, int index)
{
	DBG_ASSERT(index <= MAX_INSTANCES);
	m_instanceData.SetTrans(index, m);
}

void SubMesh::SetInstancing( std::vector<Matrix4x4> transforms )
{
	DBG_ASSERT(transforms.size() <= MAX_INSTANCES);
	for(int i=0; i<transforms.size(); i++) m_instanceData.SetTrans(i, transforms[i]);
}

void SubMesh::SetInstancing(std::vector<Matrix4x4> transforms, unsigned int startInd, unsigned int count)
{
	DBG_ASSERT(count <= MAX_INSTANCES);
	DBG_ASSERT(startInd + count <= transforms.size());
	for(int i=startInd; i<startInd+count; i++) m_instanceData.SetTrans(i-startInd, transforms[i]);
}

float SubMesh::GetExtent( int dim )
{
	DBG_ASSERT(m_loaded && dim >=0 && dim <=3);
	return m_upperBound[dim] - m_lowerBound[dim];
}

void SubMesh::AddTriangle( Vertex v1, Vertex v2, Vertex v3 )
{
	m_vertPosList.push_back(v1.position);
	m_vertPosList.push_back(v2.position);
	m_vertPosList.push_back(v3.position);
	m_normalList.push_back(v1.normal);
	m_normalList.push_back(v2.normal);
	m_normalList.push_back(v3.normal);
	m_texCoordList.push_back(v1.texCoord);
	m_texCoordList.push_back(v2.texCoord);
	m_texCoordList.push_back(v3.texCoord);
	m_indexList.push_back(m_indexList.size());
	m_indexList.push_back(m_indexList.size());
	m_indexList.push_back(m_indexList.size());
}

void SubMesh::DrawCubemap( Texture* t1, Texture* t2, float transition )
{
	if(!m_loaded) return;
	AssetManager::GetInstance()->UseProgram(AssetManager::GetInstance()->GetCubemapProg());
	glUseProgram(AssetManager::GetInstance()->GetCubemapProg()->GetId());
	glUniform1i(AssetManager::GetInstance()->GetCubemapProg()->GetVarId("cubeTex0"), 0);
	glUniform1i(AssetManager::GetInstance()->GetCubemapProg()->GetVarId("cubeTex1"), 1);
	glUniform1f(AssetManager::GetInstance()->GetCubemapProg()->GetVarId("transition"), transition);
	glActiveTexture(GL_TEXTURE0);
	glEnable(GL_TEXTURE_CUBE_MAP);
	glBindTexture(GL_TEXTURE_CUBE_MAP, t1->GetId());
	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_CUBE_MAP, t2->GetId());

	glEnable(GL_TEXTURE_GEN_S); 
	glEnable(GL_TEXTURE_GEN_T); 
	glEnable(GL_TEXTURE_GEN_R);
	glEnable(GL_NORMALIZE);

	glBindBuffer(GL_ARRAY_BUFFER, m_vertexVbo);         // for vertex coordinates
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glVertexPointer(3, GL_FLOAT, 0, 0);
	glBindBuffer(GL_ARRAY_BUFFER, m_normalVbo);
	glNormalPointer(GL_FLOAT, 0, 0);
	glBindBuffer(GL_ARRAY_BUFFER, m_vertexVbo);
	glTexCoordPointer(3, GL_FLOAT, 0, 0);
	glDrawArrays(GL_TRIANGLES, 0, m_numFaces*3);
}

void SubMesh::SaveToBin( std::ofstream& fileStrm )
{
	bool usesInstance = true;
	bool usesDeform = false;
	bool postProcess = false;
	fileStrm.write((char*)&usesInstance, sizeof(bool));
	fileStrm.write((char*)&usesDeform, sizeof(bool));
	fileStrm.write((char*)&postProcess, sizeof(bool));

	// save number of positions
	unsigned int numPositions = m_vertPosList.size();
	fileStrm.write((char*)&numPositions, sizeof(unsigned int));
	// save positions buffer
	for(int i=0; i<m_vertPosList.size(); i++) {
		for(int j=0; j<3; j++) {
			float f = m_vertPosList[i][j];
			fileStrm.write((char*)&f, sizeof(float));
		}
	}

	// save number of indices
	unsigned int numIndices = m_vertPosList.size();
	fileStrm.write((char*)&numIndices, sizeof(unsigned int));

	// save number of normals
	unsigned int numNormals = m_normalList.size();
	fileStrm.write((char*)&numNormals, sizeof(unsigned int));
	// save normals buffer
	for(int i=0; i<m_normalList.size(); i++) {
		for(int j=0; j<3; j++) {
			float f = m_normalList[i][j];
			fileStrm.write((char*)&f, sizeof(float));
		}
	}

	// save number of texcoords
	unsigned int numTexCoords = m_texCoordList.size();
	fileStrm.write((char*)&numTexCoords, sizeof(unsigned int));
	// save texcoords buffer
	for(int i=0; i<m_texCoordList.size(); i++) {
		for(int j=0; j<2; j++) {
			float f = m_texCoordList[i][j];
			fileStrm.write((char*)&f, sizeof(float));
		}
	}

	// save material info
	fileStrm.write((char*)&(m_material.m_type), sizeof(int));
	fileStrm.write((char*)&(m_material.m_col), sizeof(Point3D));

	// save bounding box
	fileStrm.write((char*)&(m_lowerBound), sizeof(Point3D));
	fileStrm.write((char*)&(m_upperBound), sizeof(Point3D));
	fileStrm.write((char*)&(m_maxExtent), sizeof(float));

	// save magic number
	fileStrm.write((char*)&END_OF_SUBMESH, sizeof(int));
}

void SubMesh::LoadFromBin( std::ifstream& fileStrm, MeshFlags flags)
{
	bool usesInstance = false;
	bool usesDeform = false;
	bool postProcess = false;
	fileStrm.read((char*)&usesInstance, sizeof(bool));
	fileStrm.read((char*)&usesDeform, sizeof(bool));
	fileStrm.read((char*)&postProcess, sizeof(bool));

	// get number of positions
	fileStrm.read((char*)&m_numPositions, sizeof(unsigned int));
	// make positions buffer
	m_positionBuf = new float[m_numPositions*3*(flags.useInstancing?MAX_INSTANCES:1)];
	fileStrm.read((char*)m_positionBuf, m_numPositions*3*sizeof(float));
	if(flags.useInstancing) { // populate rest of vertex array
		unsigned int arrInd = m_numPositions*3;
		for(int inst = 1; inst < MAX_INSTANCES; inst++) { // inst 0 is already populated
			for(int i=0; i<m_numPositions; i++) {
				for (int dim=0; dim<3; dim++) {
					m_positionBuf[arrInd] = m_positionBuf[(i*3)+dim];
					arrInd++;
				}
			}
		}
	}
	glGenBuffersARB(1, &m_vertexVbo);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, m_vertexVbo);
	glBufferDataARB( GL_ARRAY_BUFFER_ARB, m_numPositions*3*sizeof(float)*(flags.useInstancing?MAX_INSTANCES:1), m_positionBuf, GL_STATIC_DRAW_ARB);

	// get number of indices
	unsigned int numIndices;
	fileStrm.read((char*)&numIndices, sizeof(unsigned int));
	m_numFaces = numIndices/3;

	// get number of normals
	unsigned int numNormals;
	fileStrm.read((char*)&numNormals, sizeof(unsigned int));
	// save normals buffer
	m_normalBuf = new float[numNormals*3*(flags.useInstancing?MAX_INSTANCES:1)];
	fileStrm.read((char*)m_normalBuf, numNormals*3*sizeof(float));
	if(flags.useInstancing) { // populate rest of normal array
		for(int inst = 1; inst < MAX_INSTANCES; inst++) { // inst 0 is already populated
			for(int i=0; i<numNormals; i++) {
				for (int dim=0; dim<3; dim++) {
					m_normalBuf[(inst*numNormals) + (i*3) + dim] = m_normalBuf[(i*3)+dim];
				}
			}
		}
	}
	glGenBuffersARB(1, &m_normalVbo);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, m_normalVbo);
	glBufferDataARB( GL_ARRAY_BUFFER_ARB, numNormals*3*sizeof(float)*(flags.useInstancing?MAX_INSTANCES:1), m_normalBuf, GL_STATIC_DRAW_ARB);


	// get number of texcoords
	unsigned int numTexCoords;
	fileStrm.read((char*)&numTexCoords, sizeof(unsigned int));
	// make texcoords buffer
	float* texCoords = new float[numTexCoords*2*(flags.useInstancing?MAX_INSTANCES:1)];
	fileStrm.read((char*)texCoords, numTexCoords*2*sizeof(float));
	if(flags.useInstancing) { // populate rest of texcoord array
		unsigned int arrInd = numTexCoords*2;
		for(int inst = 1; inst < MAX_INSTANCES; inst++) { // inst 0 is already populated
			for(int i=0; i<numTexCoords; i++) {
				for (int dim=0; dim<2; dim++) {
					texCoords[arrInd] = texCoords[(i*2)+dim];
					arrInd++;
				}
			}
		}
	}
	glGenBuffersARB(1, &m_texCoordVbo);	
	glBindBuffer(GL_ARRAY_BUFFER, m_texCoordVbo);
	glBufferDataARB( GL_ARRAY_BUFFER_ARB, numTexCoords*2*sizeof(float)*(flags.useInstancing?MAX_INSTANCES:1), texCoords, GL_STATIC_DRAW_ARB);
	delete [] texCoords;
	texCoords = 0;

	// Make instance ids if using instancing
	glGenBuffers(1, &m_instanceIdBuf);
	glBindBuffer(GL_ARRAY_BUFFER, m_instanceIdBuf);
	if(flags.useInstancing) {
		unsigned int arrInd = 0;
		float* instanceIds = new float[m_numPositions*MAX_INSTANCES];
		for(int inst=0; inst<MAX_INSTANCES; inst++) {
			for(int i=0; i<m_numPositions; i++) {
				instanceIds[arrInd] = inst;
				arrInd++;
			}
		}
		glBufferDataARB( GL_ARRAY_BUFFER_ARB, m_numPositions*MAX_INSTANCES*sizeof(float), instanceIds, GL_STATIC_DRAW_ARB);
		delete [] instanceIds;
	}
	else { // everything has an instance id of 0
		float* instanceIds = new float[m_numPositions];
		memset(instanceIds, 0, sizeof(float) * m_numPositions);
		glBufferDataARB( GL_ARRAY_BUFFER_ARB, m_numPositions*sizeof(float), instanceIds, GL_STATIC_DRAW_ARB);
		delete [] instanceIds;
	}

	// get material info
	fileStrm.read((char*)&(m_material.m_type), sizeof(int));
	fileStrm.read((char*)&(m_material.m_col), sizeof(Point3D));

	// get bounding box
	fileStrm.read((char*)&(m_lowerBound), sizeof(Point3D));
	fileStrm.read((char*)&(m_upperBound), sizeof(Point3D));
	fileStrm.read((char*)&(m_maxExtent), sizeof(float));

	// get magic number
	int magicNum;
	fileStrm.read((char*)&magicNum, sizeof(int));
	DBG_ASSERT(magicNum == END_OF_SUBMESH);
	m_flags = flags;

	m_loaded = true;
}
SubMesh::InstanceData::InstanceData()
{
	for(int i=0; i<MAX_INSTANCES; i++) SetTrans(i, Matrix4x4());
}

void SubMesh::InstanceData::SetTrans(int i, Matrix4x4 trans)
{
	DBG_ASSERT(i>=0 && i<MAX_INSTANCES);
	for(int r=0; r<4; r++) {
		for(int c=0; c<4; c++)
			m_instanceTransBuf[i][r][c] = trans[c][r];
	}
}